/*****************************************************************************
 *                                                                           *
 * Filename:    dnsrh-support.c                                              *
 * Author:      Tabitha Tipper <t.tipper@lancaster.ac.uk>                       *
 * Created:     31/10/03                                                     *
 * Modified:    26/01/04                                                     *
 * Description: Support functions that deal with the hashing functionality   *
 *              of the DNS Response Hash.  The module should call            *
 *              initialise_hash() during its init to create the hash.  Apart *
 *              from that the only externally visible functions should be    * 
 *              used are:                                                    *
 *              dnsrh_addV6Addr()                                            *
 *              dnsrh_checkForV4()                                           *
 *              dnsrh_checkForV6()                                           *
 *                                                                           *
 * (c) 2003 Tabitha Tipper, licensed under the GNU General Public License       *
 *     See the LICENSE file for details.                                     *
 *****************************************************************************/

#ifndef dnsrh_support_c
#define dnsrh_support_c

#include <linux/slab.h>
#include <asm/semaphore.h>

#include "dnsrh.h"

#define MAXSIZE 500	/* The maximum size for the hash */


struct semaphore v4HashSem;	/* v4 Hash mutal exclusion Semaphore */
struct semaphore v6HashSem;	/* v6 Hash mutal exclusion Semaphore */
struct semaphore currV4Sem;	/* Mutex for currentV4 struct */

single_v4 currV4;		/* Stores the last addr created by the 
				   generator */

ip_pair * v6Hash[MAXSIZE];	/* The V6 hash index */
ip_pair * v4Hash[MAXSIZE];	/* The V4 hash index */

/***
 * initialise_hash performs any initialisations that need to be performed.
 * Ideally this should mean that other people can write drop in replacements
 * for the atrocious mess that I've written here and the rest of the code
 * (thats much better, honest) won't notice at all.
 *
 * It returns 0 on succes, 1 on failure
 ***/
int dnsrh_initialise_hash()
{ 
	int i = 0;

	/* initialise the semaphores */
	sema_init(&v4HashSem, 1);
	sema_init(&v6HashSem, 1);
	sema_init(&currV4Sem, 1);

	/* initialise the first addr */
	if (down_interruptible (&currV4Sem))
		return -ERESTARTSYS;

	currV4.a = 240;
	currV4.b = 0;
	currV4.c = 0;
	currV4.d = 0;

	up(&currV4Sem);

	/* Blank the hashes */
	if (down_interruptible (&v4HashSem))
		return -ERESTARTSYS;

	if (down_interruptible (&v6HashSem))
		return -ERESTARTSYS;

	for (i = 0; i < MAXSIZE; i++) { 
		v6Hash[i] = NULL;
		v4Hash[i] = NULL;
	}
	up(&v6HashSem);
	up(&v4HashSem);

	return 0;
}

/***
 * Cleans the hash properly
 *
 * FIXME: Often causes NULL dereferencing Kernel panics.... POS, rewrite.
 *
 ***/
int dnsrh_clean_hash()
{

	int i = 0;

	if (down_interruptible (&v4HashSem))
		return -ERESTARTSYS;

	if (down_interruptible (&v6HashSem))
		return -ERESTARTSYS;

	for (i = 0; i < MAXSIZE; i++) {
		ip_pair * curr = (ip_pair *) v6Hash[i];
		while (curr != NULL) { 
			ip_pair * this = curr;
			curr = (ip_pair *) this->next;
			kfree(this);
		}
		v6Hash[i] = NULL;

		curr = (ip_pair *) v4Hash[i];
		while (curr != NULL) { 
			ip_pair * this = curr;
			curr = (ip_pair *) this->next;
			kfree(this);
		}
		v4Hash[i] = NULL;

	}

	up(&v6HashSem);
	up(&v4HashSem);

	return 0;
}


/***
 * Compares two V6 parts of two ip_pair structs
 * 
 * returns 0 if they're identical and 1 if not
 ***/
int dnsrh_compareV6(ip_pair a, ip_pair b)
{
	int i = 0;

	for (i = 0; i < 16; i++)
		if (a.v6[i] != b.v6[i]) 
			return 1;

	return 0;
}

/***
 * Compares two V4 parts of two ip_pair structs
 * 
 * returns 0 if they're identical and 1 if not
 ***/
int dnsrh_compareV4(ip_pair a, ip_pair b)
{
	int i = 0;

	for (i = 0; i < 4; i++)
		if (a.v4[i] != b.v4[i]) 
			return 1;

	return 0;
}


/***
 * Accepts a char pointer to a V4 address.
 * returns a hash key in the range (0 - MAXSIZE).
 ***/
int dnsrh_hashV4Addr(ip_pair in)
{ 
	long total = 0;
	int i = 0;

	for (i = 0; i < 4; i++) { 
		total += total * in.v4[i];
	}

	return (int) total % MAXSIZE;
}

/***
 * Accepts a char pointer to a V6 address.
 * returns a hash key in the range (0 - MAXSIZE).
 ***/
int dnsrh_hashV6Addr(ip_pair in)
{
	long total = 0;
	int i = 0;

	for (i = 0; i < 16; i++) { 
		total += total * in.v6[i];
	}

	return (int) total % MAXSIZE;
}

/*** 
 * Selects the next V4 from the programs list available and returns it.
 *
 * In the event that you run out of V4s it should loop around to the beginning
 * again.
 *
 * According to RFC 3330 the "reserved for future use" block is 240.0.0.0/4,
 * This function will use the range 240.0.0.0-240.254.254.254
 *
 * If the returned address is 0.0.0.0 then an error has occured.
 ***/
single_v4 dnsrh_generateV4()
{
	single_v4 tmp;

	tmp.a = tmp.b = tmp.c = tmp.d = 0;

	if (down_interruptible (&currV4Sem)) {
		/* return -ERESTARTSYS; */
		return tmp;
	}

	currV4.d++;

	if (currV4.d == 255) { 
		currV4.d = 0;
		currV4.c++;
		if (currV4.c == 255) { 
			currV4.c = 0;
			currV4.b++;
			if (currV4.b == 255) { 
				currV4.b = 0;
				currV4.d++;
			}
		}
	}

	/* according to Pygowotsit we can do this: */
	memcpy(&tmp, &currV4, (int) sizeof(single_v4));

	up(&currV4Sem);	

	return tmp;
}

/***
 * This function accepts an ip_pair struct containing an IPv6 address.  It
 * adds it to the Hash (if its already added it does nothing).
 *
 * It modifies the struct it was given to contain the matching V4 address.
 *
 * This function returns 0 if the address was already in the hash, 1 if its
 * been freshly added to the hash, 2 if a v4 address was overwritten and
 * -ERESTARTSYS in event of a mutex error.
 ***/
int dnsrh_addV6Addr(ip_pair * in) 
{
	int hk = 0;
	ip_pair * curr = NULL;
	ip_pair * prev = NULL;
	ip_pair * newRec = NULL;	/* The new record */
	ip_pair * lastV6 = NULL;	/* add here on V6 */
	single_v4 newV4;

	/* Get the hash key */
	hk = dnsrh_hashV6Addr(*in);

	/* look in V6 hash at pointer, follow the chain looking for a match
	 * if found then copy that entry to be returned and return 0, as its 
	 * already there */
	if (down_interruptible (&v6HashSem))
		return -ERESTARTSYS;

	curr = (ip_pair *) v6Hash[hk];
	while (curr != NULL){
		if (dnsrh_compareV6(*curr, *in) == 0) {
			memcpy(in, curr, (int) sizeof(ip_pair));
			up(&v6HashSem);
			return 0;			
		}
		lastV6 = curr;
		curr = (ip_pair *) curr->next;
	}

	up(&v6HashSem);

	/* if not found need to add it to V6 and V4 hash
	 * generate a V4 address, and add it to V4 */
	newV4 = dnsrh_generateV4();

	if (newV4.a == 0 && newV4.b == 0 && newV4.c == 0 && newV4.d == 0) { 
		printk(KERN_INFO "dnsrh: generateV4 error.\n");
	}

	/* malloc a NEW RECord (newRec)
	 * put the V4 address in it 
	 * put it into the V6 hash */
	newRec = (ip_pair *) kmalloc(sizeof(ip_pair), GFP_KERNEL);
	memcpy(newRec, in, (int) sizeof(ip_pair)); /* copy V6 */
	newRec->v4[0] = newV4.a; /* copy freshly generated V4 */
	newRec->v4[1] = newV4.b;
	newRec->v4[2] = newV4.c;
	newRec->v4[3] = newV4.d;
	newRec->next = NULL;

	/* hash the new V4 address */
	hk = dnsrh_hashV4Addr(*newRec);

	/* deal with empty chain */
	if (down_interruptible (&v4HashSem))
		return -ERESTARTSYS;

	if (v4Hash[hk] == NULL) { 
		v4Hash[hk] = newRec;
	}

	else { 
		/* follow the V4 chain looking for a match, if found remove 
		 * last entry and insert new one, because v4s can be
		 * overwritten */
		prev = (ip_pair *) v4Hash[hk];
		curr = (ip_pair *) prev->next;
		while (curr != NULL) {
			if (dnsrh_compareV4(*curr, *newRec) == 0) { 
				memcpy(in, curr, (int) sizeof(ip_pair));
				newRec->next = curr->next;
				prev->next = newRec;
				kfree(curr);

				up(&v4HashSem);
				return 2;
			}
			prev = curr;
			curr = (ip_pair *) curr->next;
		}

		/* if not found append to the chain */
		prev->next = newRec;
	}
	up(&v4HashSem);

	/* Now add it to V6 hash */
	hk = dnsrh_hashV6Addr(*newRec);

	if (down_interruptible (&v6HashSem))
		return -ERESTARTSYS;

	if (NULL == lastV6) { /* Deal with empty chain */
		v6Hash[hk] = newRec;
	}
	else { 
		/* lastV6->next points to where we want to add */
		lastV6->next = newRec;
	}
	up(&v6HashSem);


	/* return the newRec after copying it over the top */
	memcpy(in, newRec, (int) sizeof(ip_pair));

	return 1;
}

/***
 * Accepts an ip_pair struct with an IPv6 address in, looks up the
 * corresponding V4 address and fills the struct.
 *
 * If there is no corresponding V4 address it fills the in.v4[x] parts with
 * 0s and returns 1.
 *
 * If there was a V4 address is returns 0 and fills the struct correctly.
 ***/
int dnsrh_checkForV4(ip_pair * in)
{
	ip_pair * curr = NULL;
	int hk = 0;

	hk = dnsrh_hashV6Addr(*in);

	if (down_interruptible (&v6HashSem))
		return -ERESTARTSYS;

	curr = v6Hash[hk];
	while (curr != NULL) { 
		if (dnsrh_compareV6(*curr, *in) == 0) { 
			memcpy(in, curr, (int) sizeof(ip_pair));
		
			up(&v6HashSem);
			return 0;
		}
		curr = (ip_pair *) curr->next;
	}

	up(&v6HashSem);

	in->v4[0] = 0;
	in->v4[1] = 0;
	in->v4[2] = 0;
	in->v4[3] = 0;

	return 1;
}

/***
 * Accepts an ip_pair struct wtih an IPv4 address in it, it looks up the
 * corresponding V6 address and fills the struct.
 *
 * If there is no corresponding V6 address it fills the in.v6[x] parts with
 * 0s and returns 1;
 *
 * If there was a V6 address it returns 0 and fills the struct correctly.
 ***/
int dnsrh_checkForV6(ip_pair * in) 
{
	ip_pair * curr = NULL;
	int hk = 0;

	/* DEBUG 
	 printk(KERN_INFO "dnsrh: cfv6: %d.%d.%d.%d\n", in->v4[0], in->v4[1], in->v4[2], in->v4[3]); */

	hk = dnsrh_hashV4Addr(*in);

	if (down_interruptible (&v4HashSem))
		return -ERESTARTSYS;

	curr = v4Hash[hk];
	while (curr != NULL) { 
		if (dnsrh_compareV4(*curr, *in) == 0) { 
			memcpy(in, curr, (int) sizeof(ip_pair));

			up(&v4HashSem);
			return 0;
		}
		curr = (ip_pair *) curr->next;
	}

	up(&v4HashSem);

	for (hk = 0; hk < 16; hk++) { 
		in->v6[hk] = 0;
	}

	return 1;
}

#endif
