/***************************************************************************
 *                                                                         *
 * Filename:    v6tv4pm.c                                                  *
 * Author:      Tabitha Tipper <t.tipper@lancaster.ac.uk>                     *
 * Created:     07/01/04                                                   *
 * Modified:    02/02/04                                                   *
 * Description: A netfilter kernel module, rewrites incoming packets to    *
 *              from IPv6 to IPv4 before putting them back on the wire.    *
 *                                                                         *
 * (c) 2004 Tabitha Tipper, licensed under the GNU General Public License     *
 *     See the LICENSE file for details.                                   *
 ***************************************************************************/

#define DEBUG_YES

/* needed ones */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>

/* For netfilter */
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv6.h>

/* For playing with skbuffs */
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <linux/ipv6.h>

/* maybe */
#include <linux/if_ether.h>

/* definitely but who knows whats in each */
#include <linux/udp.h>
#include <linux/icmp.h>
#include <net/icmp.h>
#include <net/ip.h>
#include <net/tcp.h>
#include <net/route.h>
#include <linux/netfilter_ipv4/ip_tables.h>


/* for versioning */
/*
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif  
*/

/* doesn't do much now, will later */
#include <linux/wrapper.h>

/* my own stuff */
#include "iopm.h"
#include "../dnsrh/dnsrh.h"

#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a)*65536+(b)*256+(c))
#endif

MODULE_AUTHOR("Tabitha Tipper <t.tipper@lancaster.ac.uk>");
MODULE_DESCRIPTION("A netfilter module that rewrites incoming v6 packets to v4");
MODULE_LICENSE("GPL");

/* yes I suck, returns -1 on error */
int v6tv4pm_hti(char in) 
{ 
	switch (in) { 
		case '0':
			return 0;
		case '1':
			return 1;
		case '2':
			return 2;
		case '3':
			return 3;
		case '4':
			return 4;
		case '5':
			return 5;
		case '6':
			return 6;
		case '7':
			return 7;
		case '8':
			return 8;
		case '9':
			return 9;
		case 'a':
		case 'A':
			return 10;
		case 'b':
		case 'B':
			return 11;
		case 'c':
		case 'C':
			return 12;
		case 'd':
		case 'D':
			return 13;
		case 'e':
		case 'E':
			return 14;
		case 'f':
		case 'F':
			return 15;
		default:
			return -1;
	}

	return -1;
}


/***
 * Copies a character string to an ip_pair (v6 section)
 ***/
void v6tv4pm_ctipp(char * src, ip_pair * dst) 
{ 
	int i = 0;
	int j = 0;

	for (i = 0; i < 16; i++) { 
		/* composite of two bits of char array */
		dst->v6[i] = 
			(unsigned char) ( (dst->v6[i] 
				 | v6tv4pm_hti(src[j]) ) << 4) & 0xF0;

		dst->v6[i] += 
			(unsigned char) (dst->v6[i]
				 | v6tv4pm_hti(src[j + 1])) & 0x0F;

		j += 2;
	}
}


/***
 * Takes two ip_pair structs and inserts the LOCAL_V4_x and LOCAL_V6 data into
 * them.
 ***/
void v6tv4pm_setupipp(ip_pair * v4, ip_pair * v6) {

	v4->v4[0] = LOCAL_V4_A;
	v4->v4[1] = LOCAL_V4_B;
	v4->v4[2] = LOCAL_V4_C;
	v4->v4[3] = LOCAL_V4_D;

	v6tv4pm_ctipp(LOCAL_V6, v6);
}

/***
 * v6tv4pm_ippv4cp : ip_pair v4 copy
 *
 * copies ip_pair.v4[] to from in to out.
 ***/
void v6tv4_ippv4cp(ip_pair * in, ip_pair * out) 
{
	int i = 0;

	for (i = 0; i < 4; i++)
		out->v4[i] = in->v4[i];

}

/***
 * Copies the v4 address of an ip_pair struct (in) to a __u32 to be inserted
 * in a "struct iphdr".
 ***/
void v6tv4pm_cpyv4(ip_pair * in, __u32 * out) 
{ 
	((unsigned char *) &out)[0] = in->v4[0];
	((unsigned char *) &out)[1] = in->v4[1];
	((unsigned char *) &out)[2] = in->v4[2];
	((unsigned char *) &out)[3] = in->v4[3];
}

/***
 * hdrippcmp (struct ipHDR and IP_Pair ipV6 CoMParison)
 *
 * Compares the v6 address from an ip_pair.v6[] and the v6 address from a
 * struct in6_addr.
 *
 * It returns 1 if they're the same, 0 otherwise.
 ***/
int v6tv4pm_hdrippv6cmp(ip_pair a, struct in6_addr b) 
{ 
	int i = 0;
	for (i = 0; i < 16; i++) { 
		if (a.v6[i] != b.in6_u.u6_addr8[i]) 
			return 0;
	}

	return 1;
}

/***
 * This function deals with incoming packets and is named in the v4tv6_ops
 * struct
 ***/
static unsigned int
v6tv4_hook(unsigned int hook,
		struct sk_buff **pskb,
		const struct net_device *indev,
		const struct net_device *outdev,
		int (*okfn)(struct sk_buff *))
{

	/* Get the v6 from the skbuff
	 * put it into an ip_pair
	 * look it up from the hash
	 * if 0 returned mangled the packed into an ipv4
	 * return NF_ACCEPT regardless
	 */

	if ( (*pskb)->protocol == __constant_htons(ETH_P_IPV6)) { /* IPv6 */
		ip_pair * addr = NULL;
		int err = 0;
		struct ipv6hdr * hdr = (*pskb)->nh.ipv6h;

		printk(KERN_INFO "v6tv4pm: v6 packet.\n");

		/* Get src addr */
		addr = kmalloc(sizeof(ip_pair), GFP_KERNEL);

		for (err = 0; err < 4; err++) 
			addr->v4[err] = 0;

		for (err = 0; err < 16; err++) 
			addr->v6[err] = 0;

		addr->next = NULL;

		/* This shamelessly stolen from kernel.h then modified
		 * Thanks for the code :)
		 */
		for (err = 0; err < 16; err++) 
			addr->v6[err] = ((unsigned char *) &hdr->saddr)[err];

		/* perform lookup */
		err = dnsrh_checkForV4(addr);

		/* There was a v4 addr, rewrite */
		if (0 == err) { 
			ip_pair local_v6;
			ip_pair local_v4;
			struct iphdr * v4 = NULL;

			printk(KERN_INFO "v6tv4pm: err was 0.\n");

			v4 = kmalloc(sizeof(struct iphdr), GFP_KERNEL);

			v6tv4pm_setupipp(&local_v4, &local_v6);

			/* Change skbuff protocol type */
			(*pskb)->protocol = 
				__constant_htons(ETH_P_IP);

			/* create new ipv6hdr */
			v4->ihl = (__u8) 5;
			v4->version = (__u8) 4;

			v4->tos = (__u8) 0;

			/* + size of a v6 header (40 bytes) */
			v4->tot_len = (__u16) (hdr->payload_len + 40);

			/* See Tenenbaum pg. 443 for why these are 0 */
			v4->id = (__u16) 0;
			v4->frag_off = (__u16) 0;

			v4->ttl = (__u8) hdr->hop_limit;

			v4->protocol = (__u8) 4; /* rfc 1700 */

			v4->check = (__u8) 0; /* it is, trust us */

			/* Transfer addresses over, transfer the new v4 
			 * src address from the addr we passed to the 
			 * hash.
			 *
			 * Compare the dst address from hdr to our own v6 
			 * address (LOCAL_V6), on a match write in the v4
			 * address stored in LOCAL_V4_x.
			 *
			 * On a failed check do a lookup in the hash
			 */

			v6tv4pm_cpyv4(addr, &v4->saddr);

			if (v6tv4pm_hdrippv6cmp(local_v6, hdr->daddr)) { 
				v6tv4pm_cpyv4(&local_v4, &v4->daddr);
			} 
			else { /* lookup daddr then write it in */
				ip_pair dst;
				int i = 0;

				for (i = 0; i < 16; i++) 
					dst.v6[i] = ((unsigned char *) 
							&hdr->daddr)[i];

				err = dnsrh_checkForV4(&dst);

				if (0 == err) 
					v6tv4pm_cpyv4(&dst, &v4->daddr);
				else /* backup plan */
					v6tv4pm_cpyv4(&local_v4, &v4->daddr);
			}

			#ifdef DEBUG_YES

			printk(KERN_INFO "v6tv4pm: src: %d.%d.%d.%d (%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x), dst: %d.%d.%d.%d\n (%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x)\n.", NIPQUAD(v4->saddr), NIPQUAD(v4->daddr), hdr->saddr.in6_u.u6_addr8[0],  hdr->saddr.in6_u.u6_addr8[1], hdr->saddr.in6_u.u6_addr8[2], hdr->saddr.in6_u.u6_addr8[3], hdr->saddr.in6_u.u6_addr8[4], hdr->saddr.in6_u.u6_addr8[5], hdr->saddr.in6_u.u6_addr8[6], hdr->saddr.in6_u.u6_addr8[7], hdr->saddr.in6_u.u6_addr8[8], hdr->saddr.in6_u.u6_addr8[9], hdr->saddr.in6_u.u6_addr8[10], hdr->saddr.in6_u.u6_addr8[11], hdr->saddr.in6_u.u6_addr8[12], hdr->saddr.in6_u.u6_addr8[13], hdr->saddr.in6_u.u6_addr8[14], hdr->saddr.in6_u.u6_addr8[15], hdr->saddr.in6_u.u6_addr8[0], hdr->saddr.in6_u.u6_addr8[1], hdr->saddr.in6_u.u6_addr8[2], hdr->saddr.in6_u.u6_addr8[3], hdr->saddr.in6_u.u6_addr8[4], hdr->saddr.in6_u.u6_addr8[5], hdr->saddr.in6_u.u6_addr8[6], hdr->saddr.in6_u.u6_addr8[7], hdr->saddr.in6_u.u6_addr8[8], hdr->saddr.in6_u.u6_addr8[9], hdr->saddr.in6_u.u6_addr8[10], hdr->saddr.in6_u.u6_addr8[11], hdr->saddr.in6_u.u6_addr8[12], hdr->saddr.in6_u.u6_addr8[13], hdr->saddr.in6_u.u6_addr8[14], hdr->saddr.in6_u.u6_addr8[15]);

			#endif


			/* setup new pointer */
			kfree((*pskb)->nh.ipv6h);
			(*pskb)->nh.iph = v4;
		} 
		/* let modified packet continue */
		kfree(addr);
	}

	return NF_ACCEPT;
}

/***
 * This struct is used to register our interest for netfilter work
 ***/
static struct nf_hook_ops v6tv4_ops = 
{ { NULL, NULL }, v6tv4_hook, PF_INET6, NF_IP6_PRE_ROUTING, 0 };

/***
 * This is the function that kickstarts the module.
 ***/
static int __init v6tv4_init(void) 
{
	int err = 0;


	/* register with NF_IP_PRE_ROUTING */
	err = nf_register_hook(&v6tv4_ops);

	printk(KERN_INFO "v6tv4: IPv6 to IPv4 (incoming) filter loaded (%d).\n"
			, err); 
	return err;
}

/***
 * This function stops the module.
 ***/
static void __exit v6tv4_exit(void) 
{ 
	/* unregister with NF_IP_PRE_ROUTING */
	nf_unregister_hook(&v6tv4_ops);
	printk(KERN_INFO "v6tv4: IPv6 to IPv4 (incoming) filter unloaded.\n");
}


	module_init(v6tv4_init)
module_exit(v6tv4_exit)
